@@ -10,8 +10,10 @@ import re
 import subprocess
 import sys
 import time
+import json
 import warnings
 from collections import defaultdict
+from configparser import ConfigParser
 
 import requests
 
@@ -25,6 +27,11 @@ try:
 except ImportError:
     argcomplete = None
 
+try:
+    from configupdater import ConfigUpdater
+except ImportError:
+    ConfigUpdater = None
+
 import copr.exceptions as copr_exceptions
 from copr.v3 import (
     Client, config_from_file, CoprException, CoprRequestException,
@@ -290,6 +297,51 @@ class Commands(object):
             log.error("Can't detect who are you.")
             raise
 
+    def action_new_api_token(self, args):
+        """
+        Simply print out the current user as defined in copr config.
+        """
+
+        if not os.path.exists(args.config):
+            # We don't want to generate the file when user authenticates
+            # via gssapi, or?
+            raise CoprConfigException(f"File {args.config} not found.  Can't update.")
+
+        # We know there's a TOCTOU.  It is better to check that the file is
+        # writable *before* we regenerate the token, avoiding a situation where
+        # we have no way to write the new token.
+        if not os.access(args.config, os.W_OK):
+            raise CoprConfigException(f"File {args.config} is not writable.  Can't update.")
+
+        if not (self.config["login"] and self.config["token"]):
+            # users that use gssapi do not have token/login specified
+            raise CoprConfigException(f"Current {args.config} doesn't specify login/token.  Can't update.")
+
+        result = self.client.base_proxy.new_api_token()
+
+        self._update_config(args.config, {
+            "login": result.api_login,
+            "token": result.api_token,
+        })
+        print(f"New API token was updated in {args.config}.")
+
+
+    def _update_config(self, path, data):
+        if ConfigUpdater:
+            config = ConfigUpdater()
+            config.read(path)
+            for key, value in data.items():
+                config["copr-cli"][key].value = value
+            config.update_file()
+        else:
+            # This breaks the previous formatting, removes comments inside of
+            # the config file, etc. That's why we only use it as a fallback.
+            self.client.config.update(data)
+            config = ConfigParser()
+            config["copr-cli"] = self.client.config
+            with open(path, "w") as fp:
+                config.write(fp)
+
     def action_new_webhook_secret(self, args):
         """
         Regenerate webhook secret for a copr.
@@ -638,6 +690,10 @@ class Commands(object):
         build = self.client.build_proxy.get(args.build_id)
         base_len = len(os.path.split(build.repo_url))
         build_chroots = self.client.build_chroot_proxy.get_list(args.build_id)
+        project = self.client.project_proxy.get(
+            build.ownername,
+            build.projectname,
+        )
 
         for chroot in build_chroots:
             if args.chroots and chroot.name not in args.chroots:
@@ -647,13 +703,19 @@ class Commands(object):
                 sys.stderr.write("No data for build id: {} and chroot: {}.\n".format(args.build_id, chroot.name))
                 continue
 
+            dst = os.path.join(args.dest, chroot.name)
             cmd = ['wget', '-r', '-nH', '--no-parent', '--reject', '"index.html*"', '-e', 'robots=off', '--no-verbose']
-            cmd.extend(['-P', os.path.join(args.dest, chroot.name)])
+            cmd.extend(['-P', dst])
             cmd.extend(['--cut-dirs', str(base_len + 4)])
 
             if args.rpms:
                 cmd.extend(["-A", "*.rpm"])
 
+            # Unfortunately, when the project is in Pulp, we need also
+            # `results.json` so that we know what filenames to download from Pulp
+            if args.rpms and project.storage == "pulp":
+                cmd.extend(["-A", "results.json"])
+
             if args.spec:
                 cmd.extend(["-A", "*.spec"])
 
@@ -676,6 +738,43 @@ class Commands(object):
             cmd.append(chroot.result_url)
             subprocess.call(cmd)
 
+            if project.storage != "pulp":
+                continue
+
+            results_json = os.path.join(dst, "results.json")
+            try:
+                with open(results_json, "r", encoding="utf-8") as fp:
+                    results = json.load(fp)
+            except FileNotFoundError:
+                # The file doesn't exist when for example `--logs`, or `--spec`
+                # were specified, and in that case we can safely continue. When
+                # no filters were specified or `--rpms` was used, then the file
+                # exists.
+                continue
+
+            for nevra in results["packages"]:
+                filename = "{N}-{V}-{R}.{A}.rpm".format(
+                    N=nevra["name"],
+                    V=nevra["version"],
+                    R=nevra["release"],
+                    A=nevra["arch"],
+                )
+                url = "{baseurl}/{chroot}/Packages/{letter}/{filename}".format(
+                    baseurl=build.repo_url,
+                    chroot=chroot.name,
+                    letter=filename[0],
+                    filename=filename,
+                )
+                response = requests.get(url)
+                if not response.ok:
+                    sys.stderr.write("Failed to download: {0}\n".format(url))
+                    # Everything should be possible to download. But if not,
+                    # it is better for the user to download at least what we
+                    # can instead of immediately exiting.
+                    continue
+                with open(os.path.join(dst, filename), "wb") as fp:
+                    fp.write(response.content)
+
     @requires_api_auth
     def action_cancel(self, args):
         """ Method called when the 'cancel' action has been selected by the
@@ -1028,6 +1127,8 @@ def create_and_modify_common_opts(parser):
         metavar="FORGE_PROJECT", action="append", help=(
             "Forge project that will be allowed to build in this project "
             "via Packit in format github.com/packit/ogr. "
+            "Supports wildcard patterns (e.g., github.com/theproject/* will allow "
+            "all repositories in the theproject organization). "
             "Can be specified multiple times."
     ))
 
@@ -1054,6 +1155,11 @@ def setup_parser():
 
     subparsers = parser.add_subparsers(title="actions")
 
+    parser_new_api_token = subparsers.add_parser(
+        "new-api-token",
+        help="Generate a new API token. Be careful, this destroys the old one!")
+    parser_new_api_token.set_defaults(func="action_new_api_token")
+
     #########################################################
     ###                    Project options                ###
     #########################################################
